#include<unistd.h>
#include<errno.h>
#include<stdio.h>
#include<stdlib.h>
/* #define _REENTRANT */    /* basic 3-lines for threads */
#include <pthread.h>
/* #include <thread.h>  */
#include<time.h>


void err_abort(const int i){
 printf("%d:%s\n",i,strerror(errno));
 exit(-1);
}

int total, loopmax;

int val = 0;

typedef 
 struct channel{
   pthread_mutex_t *cin, *cout;
   int val;
   pthread_t pid;
 } channel_t;

void * from(void *arg){
 channel_t *ch = (channel_t *) arg;
 for (;;){
  pthread_mutex_lock(ch->cin);
  ch->val++;
  pthread_mutex_unlock(ch->cout);
 }
}
 
channel_t * from_create(int n){
 channel_t *ch = (channel_t *) malloc(sizeof(channel_t));
 ch->cin=(pthread_mutex_t *) malloc(sizeof(pthread_mutex_t));
 ch->cout=(pthread_mutex_t *) malloc(sizeof(pthread_mutex_t));
 pthread_mutex_init(ch->cin,NULL);
 pthread_mutex_init(ch->cout,NULL);
 ch->val = n;

 pthread_mutex_lock(ch->cin);

 if (pthread_create(&ch->pid,NULL,&from,ch) !=0)
  exit(-1);
 
 return ch;

}

typedef 
 struct filter_triple{
  int n;
  channel_t * inch, *outch;
 } filter_triple_t;
 
void * filter(void * arg){
 filter_triple_t * f=(filter_triple_t *) arg;
 channel_t *in = f->inch,
           *out = f->outch;
 int n = f->n;

 for (;;){
   int aux;
   do{
    pthread_mutex_lock(in->cout);
    aux = in->val;
    pthread_mutex_unlock(in->cin);
   } while ((aux%n)==0);
   pthread_mutex_lock(out->cin);
   out->val=aux;
   pthread_mutex_unlock(out->cout);
 }
}

channel_t *filter_create (int n, channel_t *inch){
 filter_triple_t *f = (filter_triple_t *) malloc(sizeof(filter_triple_t));

 channel_t *outch = (channel_t *) malloc(sizeof(channel_t));
 outch->cin=(pthread_mutex_t *) malloc(sizeof(pthread_mutex_t));
 outch->cout=(pthread_mutex_t *) malloc(sizeof(pthread_mutex_t));
 pthread_mutex_init(outch->cin,NULL);
 pthread_mutex_init(outch->cout,NULL);
 pthread_mutex_lock(outch->cout);
 f->inch=inch;
 f->outch=outch;
 f->n=n;

 if (pthread_create(&f->outch->pid,NULL,&filter,f) !=0)
  exit(-1);
 
 return outch;
}


typedef 
 struct sieve_pair{
  channel_t * inch, *outch;
 } sieve_pair_t;


void *sieve(void *arg){
 sieve_pair_t *s = (sieve_pair_t*) arg;
 channel_t * in = s->inch;
 channel_t * out = s->outch;

 for(;;){
  int hd;
  pthread_mutex_lock(in->cout);
  hd = in->val;
  pthread_mutex_unlock(in->cin);
  pthread_mutex_lock(out->cin);
  out->val=hd;
  pthread_mutex_unlock(out->cout);
  in = filter_create(hd,in);
 }
}


channel_t * sieve_create(channel_t *inch){
 sieve_pair_t *s = (sieve_pair_t *) malloc(sizeof(sieve_pair_t));

 channel_t *outch = (channel_t *) malloc(sizeof(channel_t));
 outch->cin=(pthread_mutex_t *) malloc(sizeof(pthread_mutex_t));
 outch->cout=(pthread_mutex_t *) malloc(sizeof(pthread_mutex_t));
 pthread_mutex_init(outch->cin,NULL);
 pthread_mutex_init(outch->cout,NULL);
 pthread_mutex_lock(outch->cout);
 s->inch=inch;
 s->outch=outch;

 if (pthread_create(&outch->pid,NULL,&sieve,s) !=0)
  exit(-1);
 
 return outch;
}




int take (channel_t *ch){
 int v;
 pthread_mutex_lock(ch->cout);
 v = ch->val;
 pthread_mutex_unlock(ch->cin);
 return v;
}



 
int main(int argc, char **argv){
 int i,init,n;
 channel_t *ch;
 pthread_mutex_t m;

 sscanf(argv[1],"%d",&init);
 sscanf(argv[2],"%d",&n);

 pthread_mutex_init(&m,NULL);
 ch = sieve_create(from_create(init));
 for (i=0;i<n;i++){
  int v = take(ch);
  printf("v:%d\n",v);

 }
 

 exit(0);

}


