/*
 * Laboratory for Scientific Computing
 * http://www.lam-mpi.org/tutorials/
 * University of Notre Dame
 *
 * MPI Tutorial
 * Homework: Manager/worker (calculate an average in parallel)
 *
 * Mail questions regarding tutorial material to mpi@lam-mpi.org
 */

#include "mpi.h"
#include "stdio.h"
#include <malloc.h>

int
main(int argc, char *argv[])
{
  int rank, size;
  int i, len = 100, tag = 200;

  MPI_Init(&argc, &argv);   

  MPI_Comm_rank(MPI_COMM_WORLD, &rank);
  MPI_Comm_size(MPI_COMM_WORLD, &size);

  if (rank == 0) {                            /* Manager does the following */

    /* Allocate and initialize the data */ 

    int total, workers = size - 1;
    double avg;
    int *data = malloc(sizeof(int) * workers * len);
    int *sum = malloc(sizeof(int) * workers);
    MPI_Request *request = malloc(sizeof(MPI_Request) * workers);
    MPI_Status *status = malloc(sizeof(MPI_Status) * workers);

    for (i = 0; i < workers * len; i++)
      data[i] = i;
    
    /* Send the data to each of the other processes */

    for (i = 0; i< workers; i++) {
      int dest = i + 1;
    
    /* Do any immediate send to all the workers and give them their numbers */  
      MPI_Isend(data + i * len, len, MPI_INT, dest, tag+dest,
		MPI_COMM_WORLD, &request[i]);
    }
    MPI_Waitall(workers, request, status);
    
    /* Receive the results */

    for (i = 0; i < workers; i++) {
      int source = i + 1;
      MPI_Irecv(sum + i, 1, MPI_INT, source, tag + source, 
		MPI_COMM_WORLD, &request[i]);
    }
    MPI_Waitall(workers, request, status);

    total = 0;
    for (i = 0; i < workers; i++)
      total += sum[i];

    avg = total / (len * workers);
    printf("The average is %lf\n", avg);

  } else {
    int data[100], sum = 0;
    MPI_Status status;

    MPI_Recv(data, len, MPI_INT, 0, tag + rank, MPI_COMM_WORLD, &status);
    for (i = 0; i < len; i++)
      sum += data[i];
    
    MPI_Send(&sum, 1, MPI_INT, 0, tag + rank, MPI_COMM_WORLD);

  }

  MPI_Finalize();
  return 0;
}
